open Options
open Debug_options

module Sy = Symbols
module T = Term
module TS = T.Set
module TM = T.Map
module MapS = Sy.Map
module F = Format

type pair = T.t * T.t
type proof = In of pair | Congr of pair

let symb t = let {T.f = f; xs = xs } = T.view t in f
let sub_list t = let {T.f = f; xs = xs } = T.view t in xs

let interpreted f= 
  if Sy.equal f Sy.plus || Sy.equal f Sy.minus then true
  else
    match f with | Sy.Int _ -> true | _ -> false

let rec print_list sep printer fmt = function
    [] -> ()
  | [x] -> F.fprintf fmt "%a" printer x
  | (x::l) -> F.fprintf fmt "%a%s%a" printer x sep (print_list sep printer) l

let print_coq_list printer fmt l =
  if l = [] then F.fprintf fmt "nil"
  else F.fprintf fmt "(%a :: nil)" (print_list " :: " printer) l

let print_symbol_list fmt sym =
  let keys_to_list m =
    List.rev (MapS.fold (fun k _ acc -> k::acc) m [] )
  in
    F.fprintf fmt "Inductive symb2 : Set := %a.@," 
      (print_list "|" Sy.print) (keys_to_list sym)
  
let print_symbol_ord fmt sym =
  let a = ref 0 in
    MapS.iter 
      (fun k _ -> 
         F.fprintf fmt "| %a => %d@\n" Sy.print k !a;
         incr a) sym

let print_import fmt s = F.fprintf fmt "Require Import %s.@\n" s
let print_modules fmt l = List.iter (print_import fmt) l

let print_sigmodule fmt sym =
  print_modules fmt ["Relations"; "List"; "listdec"; "o_terms";
                     "listrel"; "cc"; "arith"];
  F.fprintf fmt "@[<hov 2>Module Mysig <: OrdSig.@\n";
  F.fprintf fmt "Ltac fresh_case := let name := fresh in intro name; case name.@\n";
  print_symbol_list fmt sym;
  F.fprintf fmt "Definition symb := symb2.@\n";
  F.fprintf fmt "Definition A := symb.@\n";
  F.fprintf fmt "Definition eq_symbol_dec : forall f1 f2 : symb, {f1 = f2} + {f1 <> f2}.@\n";
  F.fprintf fmt "Proof. decide equality. Defined.@\n";
  F.fprintf fmt "Inductive arity_type : Set := AC : arity_type | C : arity_type | Free : nat -> arity_type.@\n";
  F.fprintf fmt "@[<hov 2>Definition arity (f:symb) :=@\n";
  F.fprintf fmt "@[<hov 2>match f with@\n";
  MapS.iter (fun k v -> F.fprintf fmt "| %a => Free %d@\n" Sy.print k v) sym;
  F.fprintf fmt "end.@]@]@\n";
  F.fprintf fmt "@[<hov 2>Definition ord_f (f:symb) : nat :=@\n";
  F.fprintf fmt "@[<hov 2>match f with@\n";
  print_symbol_ord fmt sym;
  F.fprintf fmt "end.@]@]@\n";
  F.fprintf fmt "Definition ord_f_dec : forall (a b: A), {ord_f a <= ord_f b} + {ord_f a > ord_f b}.@\n";
  F.fprintf fmt "Proof. do 2 fresh_case; simpl; solve [ left; auto with arith | right ; auto with arith]. Defined.@\n";
  F.fprintf fmt "@[<hov 2>Definition o_A : relation A := fun (a b:A) =>@\n";
  F.fprintf fmt "if ord_f_dec a b then True else False.@]@\n";
  F.fprintf fmt "Lemma o_A_proof : order A o_A.@\n";
  F.fprintf fmt "@[<hov 2>Proof.@\n";
  F.fprintf fmt "constructor; [fresh_case | do 3 fresh_case | do 2 fresh_case ];@\n";
  F.fprintf fmt "unfold o_A; simpl; intros; try contradiction; auto.@]@\n";
  F.fprintf fmt "Qed.@\n";
  F.fprintf fmt "Lemma o_A_dec : forall e1 e2 : A, {o_A e1 e2} + {~ o_A e1 e2}.@\n";
  F.fprintf fmt "Proof. do 2 fresh_case; unfold o_A; simpl; solve [left; auto | right; auto]. Defined.@\n";
  F.fprintf fmt "Lemma o_A_total : forall e1 e2 : A, {o_A e1 e2} + {o_A e2 e1}.@\n";
  F.fprintf fmt "  Proof. do 2 fresh_case; unfold o_A; simpl; solve [left; auto | right; auto]. Defined.@]@\n";
  F.fprintf fmt "End Mysig.@\n@\n"

let print_inst fmt =
  F.fprintf fmt "Module MyA := arith.Arithmetics Mysig.@\n";
  F.fprintf fmt "Module EmCC := CC MyA.@\n";
  F.fprintf fmt "Import Mysig. Import EmCC.T. Import EmCC. Import MyA.T.Aordsig.@\n";
  F.fprintf fmt "Definition inc := in_correct (pair_dec term eq_term_dec).@\n";
  F.fprintf fmt "Definition lrc := fun (u:Uf.t) => listrel_correct term (Uf.equal u) (Uf.uf_eq_dec u).@\n";
  F.fprintf fmt "Definition rt := refl_equal true.@\n";
  F.fprintf fmt "Ltac lazy_generic := lazy beta delta iota zeta; exact rt.@\n";
  F.fprintf fmt "Ltac lazy_repr proof t1 t2 :=@\n";
  F.fprintf fmt "  let name := fresh in@\n";
  F.fprintf fmt "  assert (name : Uf.uf_ref (get_u proof) t1 t2 = true) by lazy_generic;\n";
  F.fprintf fmt "  apply correct_cc with (get_u proof) (get_o proof); @\n";
  F.fprintf fmt "    [ exact proof| apply Uf.uf_dec_correct; exact name].@\n";
  F.fprintf fmt "Ltac lazy_congr fs C U l1 l2:=";
  F.fprintf fmt "  let name := fresh in@\n";
  F.fprintf fmt "  assert (name : lr_decb term (Uf.equal U) (Uf.uf_eq_dec U) l1 l2 = true)@\n";
  F.fprintf fmt "  by lazy_generic; exact (congr (UIF fs) C (lrc U l1 l2 name)).@\n"

let prsh map fmt t = F.fprintf fmt "t%d" (TM.find t map)
let rec prhl map fmt l = print_coq_list (prsh map) fmt l

let symb_print fmt f =
  if interpreted f then
    if Sy.equal f Sy.plus then
      F.fprintf fmt "INF plus"
    else
      if Sy.equal f Sy.minus then
        F.fprintf fmt "INF minus"
      else
        match f with
          | Sy.Int k ->
              F.fprintf fmt "INF (const %d)" (int_of_string (Hstring.view k))
          | _ -> assert false
  else
    F.fprintf fmt "UIF %a" Sy.print f

let prh map fmt t =
  F.fprintf fmt "Term (%a) %a" symb_print (symb t) (prhl map) (sub_list t)
        
let print_terms fmt tl tm =
  F.fprintf fmt "@[<hov 2>Module T1.@\n";
  List.iter 
    (fun (k,v) -> 
      F.fprintf fmt "Definition t%d := %a. (* %a *) @\n" v (prh tm) k T.print k)
    tl 

let print_equ fmt tm l =
  let print_pair fmt (t1,t2) =
    let psm = prsh tm in
      F.fprintf fmt "(%a,%a)" psm t1 psm t2
  in
  F.fprintf fmt "@\nDefinition e := %a.@\n" (print_coq_list print_pair) l

let print_ufs fmt tm cl =
  let pt = prsh tm in
  let to_pair = function
    | In (t1,t2) | Congr (t1,t2) -> (t1,t2) in
  let aux i t1 t2 =
    F.fprintf fmt "Definition U%d := Uf.union U%d %a %a.@\n" i (i-1) pt t1 pt t2
  in
  let rec aux2 i = function
    | [] -> ()
    | ((t1,t2)::xs) -> begin aux i t1 t2; aux2 (i+1) xs end
  in
    F.fprintf fmt "Definition U0 := Uf.empty.@\n";
    aux2 1 (List.map to_pair cl)

let print_proof fmt tm cl =
  let pt = prsh tm in
  let print_term_list fmt l = print_coq_list pt fmt l in
  let print_pair fmt (t1,t2) =
      F.fprintf fmt "(%a,%a)" pt t1 pt t2 in
  let to_pair = function
    | In (t1,t2) | Congr (t1,t2) -> (t1,t2) in
  let prtype i fmt acc = 
    F.fprintf fmt "Conf e %a U%d" (print_coq_list print_pair) (List.map to_pair acc) i in
  let print_single_proof i fmt = function
    | In (t1,t2) ->
        F.fprintf fmt "exact (in_p %a %a C%d (inc (%a,%a) e rt)). Qed." 
          pt t1 pt t2 i pt t1 pt t2
    | Congr (t1,t2) ->
        F.fprintf fmt "lazy_congr %a C%d U%d %a %a. Qed." Sy.print (symb t1) i
          i print_term_list (sub_list t1) print_term_list (sub_list t2)
  in
  let rec iter_proofs i fmt acc = function
    | [] -> i - 1
    | (x::xs) -> 
        let acc = x::acc in
          F.fprintf fmt "Definition C%d : %a.@\n%a@\n" i (prtype i) acc 
            (print_single_proof (i-1)) x;
          iter_proofs (i+1) fmt acc xs
  in
    F.fprintf fmt "Definition C0 : Conf e nil U0 := init e.@\n";
    iter_proofs 1 fmt [] cl
  
let print_lemma fmt tm t1 t2 i =
  let pt = prsh tm in
  F.fprintf fmt "Lemma proof : MyA.thex e %a %a.@\n" pt t1 pt t2;
  F.fprintf fmt "Proof. lazy_repr C%d %a %a. Qed.@\n" i pt t1 pt t2

let trace = ref (MapS.empty,[],[], TS.empty)

let rec arity t m =
  let {T.f = f; xs = xs } = T.view t in
  let m =
    if MapS.mem f m || interpreted f then m
    else MapS.add f (List.length xs) m 
  in
    List.fold_left (fun acc x -> arity x acc) m xs

let add_eq t1 t2 ccb =
  let m,cl,eql,s = !trace in
  let m = arity t1 (arity t2 m) in
  let cl,eql =  
    if ccb then (Congr (t1,t2)::cl), eql
    else (In (t1,t2)::cl), ((t1,t2)::eql)
  in
    trace := m,cl,eql, TS.add t1 (TS.add t2 s)

module TermGraph =
struct
  type t = int TM.t
  module V = T
  let create s =
    let add_prot k m = 
      let i = try TM.find k m with Not_found -> 0 in TM.add k (i+1) m
    in
    let add_zero k m = if TM.mem k m then m else TM.add k 0 m in
    let rec add_rec t (m, s) =
      if TS.mem t s then m,s
      else
        let m = add_zero t m in
        let s = TS.add t s in
        let xs = sub_list t in
        let m = List.fold_left (fun acc st -> add_prot st acc) m xs in
          List.fold_left (fun acc t -> add_rec t acc) (m,s) xs
    in
    let m,s = TS.fold (fun t acc -> add_rec t acc) s (TM.empty, TS.empty) in
      m

  let iter_succ f s t =
    let {T.f = f1; xs = xs } = T.view t in
      List.iter f xs

  let in_degree m t = try TM.find t m with Not_found -> 0
  let iter_vertex f s = TM.iter (fun k _ -> f k) s
end

module Top = Graph.Topological.Make (TermGraph)

let build_names m = 
  let i = ref (-1) in
    Top.fold (fun x acc -> (incr i; (x,!i)::acc)) m [] 

let build_map tl =
    List.fold_left (fun acc (k,v) -> TM.add k v acc) TM.empty tl

let print_trace t1 t2 =
  let cout = open_out tracefile in
  let fmt = F.formatter_of_out_channel cout in
(*   let fmt = F.std_formatter in *)
  let sym, cl,eql,s = !trace in
  let s = TS.add t1 (TS.add t2 s) in
  let () = print_sigmodule fmt sym in
  let () = print_inst fmt in
  let m = TermGraph.create s in
  let tl = build_names m in
  let tm = build_map tl in
  let () = print_terms fmt tl tm in
  let () = print_equ fmt tm eql in
  let cl = List.rev cl in
  let () = print_ufs fmt tm cl in
  let i = print_proof fmt tm cl in
  let () = print_lemma fmt tm t1 t2 i in
    F.fprintf fmt "@]@\nEnd T1.@\n@.";
    close_out cout
